/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "delay.h"
#include "flag.h"
#define clight  2.9979250e8 /*m per sec */
#define  DELTAOPDPRISM 150.e-6
#define  DELTAOPDGRISM 800.e-6
#define  JUMPOPD     10.e-6
#define  JITTEROPD   1.5e-6 /* max allowed opd jitter */
#define  JITTERBOX 50
#define  JITTERDD  1
#ifdef HP
#define finite(x) isfinite(x)
#endif


void Usage() {
   printf("   oirAutoFlag inFile delayFile flagFile [-maxopd maxopd] [-minopd minopd] [-jump jumpopd] [-jitteropd jitteropd] [-SN sigma\n");
   printf("      inFile = file name of output of oirRotateGroupDelay\n");
   printf("      delayFile  = file name of output of oirGroupDelay\n");
   printf("      flagFile   = file name of output file containing flag data\n");
   printf("      maxopd     = maximum allowed difference between tracking and true OPD (microns)\n");
   printf("        default  = 150 for PRISM and 800 for GRISM\n");
   printf("      minopd     = minimum allowed difference between tracking and true OPD (microns)\n");
   printf("         default = 0 \n");
   printf("      jumpopd    = maximum allowed jump in opd between frames (microns, default = 10\n");
   printf("      jitteropd  = maximum acceptable short term rms opd before flagging (microns, default = 1.5\n");
   printf("      sigma      = minimum allowed signal/noise;default: no SN flagging \n");
   return;
}

/********************************************************************
 * oirAutoFlag "flags" (i.e. sets a flag marking a point as invalid)
 * those points where the group delay differs from the tracked
 * value of the delay by more than the value of the parameter
 * "maxopd" or less than the value of the parameter "minopd".
 *
 * It also flags points where the group delay jumps more than
 * "jumpOpd" as measured by the absolute value of the 2nd 
 * difference of the group Delay
 ************************************************************************/

float quadInterpol(int nData, double *x, float *y, double t, int *iLast) {

   int i;
   float a,b,c,d,dx;
   float xmin, xmax, dmin, dmax;

   while ((t > x[*iLast]) && ((*iLast) < (nData-1))) (*iLast)++;
   i = *iLast;
   if (i == 0) return y[0];
   if (i == (nData-1)) return y[nData-1];

   xmin = x[i]-x[i-1];
   xmax = x[i+1]-x[i];
   dmin = y[i] - y[i-1];
   dmax = y[i+1] - y[i];
   dx = t - x[i];
   d = xmin * xmax * (xmin+xmax);
   a = y[i];
   b = (dmax * xmin * xmin + dmin * xmax *xmax)/d;
   c = (dmax * xmin - dmin * xmax)/d;
   return a + dx*(b + c *dx);
}

void minrtsMedianSmooth(float *data, int nData, int nSmooth, float* outData) {
   float *mData;
   int nOdd, n2;
   int iMin, iMax, i;
   int out, odd;

   odd = nSmooth - 2 * (nSmooth/2);
   if (odd) nOdd = nSmooth; else nOdd = nSmooth+1;
   n2 = (int)(nOdd/2);
   mData = (float *)malloc(nOdd * sizeof(float));

   for (out=0; out<nData; out++) {
      iMin = out - n2;
      if (iMin < 0) iMin = 0;
      iMax = out + n2;
      if (iMax >= nData) iMax = nData-1;
      for (i=iMin;i<=iMax;i++) mData[i-iMin] = data[i];
      minrtsHeapSort(1+iMax-iMin, mData);
      outData[out] = mData[n2];
   }
   free(mData);
   return;
}


int main (int argc, char** argv) {

   int i, j, ierr;
   long iRow, sRow, oRow, pRow, dRow ;
   long  nDelay, dataRows;
   int nx;

   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   int iKey ;
   double opd;
   imaging_fdata_table_def *idata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *inData;

   delay_table_row *delayData;
   delay_table_def *dTable;
   double *delayTimes;
   float *groupDelays, *medianGroupDelay, *medianJitter, *jOpd;
   double startDataTime, startDelayTime;
   float  *delayAmp;
   double  *dataTimes;

   float maxOpd, minOpd, jumpOpd;
   float jitterOpd;
   int   jitterDd;
   float dOpd ;
   long nJump, nMax, nGood, nAmp, nJit, nNan;

   char gris[12];
   char history[80];

   flag_table_row *flagData;
   flag_table_def *flagTable;
   int *dFlags, *jFlags, *aFlags, *nanFlags, 
      *rFlags;
   int nRflag;
   float rmsRflag;
   int *ImOK;

   float intOpd;
   double inTime;
   int   xLast;
   double time0, time1;

   float dNoise;
   float snLevel;

   int iarg;

   ierr   = 0;
   nJump  = 0;
   nNan   = 0;
   nMax   = 0;
   nAmp   = 0;
   nJit   = 0;
/*
      parse input; first set defaults then read command line
*/
   maxOpd     = 0.;
   minOpd     = 0.;
   jumpOpd    = JUMPOPD;
   jitterOpd  = JITTEROPD;
   jitterDd   = JITTERDD;

   snLevel  = -1; /* no SN flagging by default */

   iarg = 4;
   while (iarg < argc ) {
      if (!strcmp("-jump", argv[iarg])) {
         sscanf(argv[++iarg],"%f",&jumpOpd);
         jumpOpd = jumpOpd * 1.e-6; /* microns to meters */
      }
      if (!strcmp("-maxopd", argv[iarg])) {
         sscanf(argv[++iarg],"%f",&maxOpd);
         maxOpd = maxOpd * 1.e-6; /* microns to meters */
      }
      if (!strcmp("-minopd", argv[iarg])) {
         sscanf(argv[++iarg],"%f",&minOpd);
         minOpd = minOpd * 1.e-6; /* microns to meters */
      }
      if (!strcmp("-jitteropd", argv[iarg])) {
         sscanf(argv[++iarg],"%f",&jitterOpd);
         jitterOpd = jitterOpd * 1.e-6; /* microns to meters */
      }
      if (!strcmp("-jitterdd", argv[iarg])) {
         sscanf(argv[++iarg],"%i",&jitterDd);
      }
      if (!strcmp("-SN", argv[iarg]))   sscanf(argv[++iarg],"%f",&snLevel);
      ++iarg;
   }
   if (argc < 3) {
      Usage();
      return 1;
   }
   printf ("      ******   oirAutoFlag   ******\n");
   printf ("   Using dataFile %s and delFile %s\n", argv[1], argv[2]);
   printf ("   Using output File %s\n", argv[3]);

   if (jitterOpd < 0.) jitterOpd = 1.e-3;
/*
      open delay file and read data into memory
*/
   fits_open_file (&inFile, argv[2], READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits open delay file of %s gave error %i\n",argv[2],
         ierr);
      printf("%s .\n",strerror(errno));
      return 1;
   }
   dTable = open_delay_table(inFile, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits open delay table error %i\n", ierr);
      return 1;
   }
   i = fits_read_key(inFile, TFLOAT, "DNOISE", &dNoise, 0, &ierr);
   if (ierr != 0) {
/*      printf("   error reading dNoise from input delay file %i.\n", ierr);
      printf("   ignoring signal/Noise flagging\n"); */
      dNoise = -1.;
      ierr   = 0;
   } else printf("   input delay amplitude noise is %f.\n", dNoise);
   if ((snLevel <= 0) || (dNoise <= 0.))  printf ("   No SN flagging specified.\n");
      else printf ("SN flagging specified at S/N = %f.\n", snLevel);

   i = fits_get_num_rows (inFile,  &nDelay, &ierr); 
   delayData = make_delay_table_row(dTable);
/*
         set up memory arrays
*/
   delayTimes  = (double *)malloc(nDelay * sizeof(double));
   groupDelays = (float *)malloc(nDelay * sizeof(double));
   medianGroupDelay = (float *)malloc(nDelay * sizeof(double));
   medianJitter     = (float *)malloc(nDelay * sizeof(double));
   jOpd        = (float *)malloc(nDelay * sizeof(int));
   delayAmp    = (float *)malloc(nDelay * sizeof(float));
   jFlags      = (int *)malloc(nDelay * sizeof(int));
   rFlags      = (int *)malloc(nDelay * sizeof(int));
   aFlags      = (int *)malloc(nDelay * sizeof(int));


/*
         read delay data
*/
   for (iRow =1; iRow<=nDelay; iRow++) {
      i = read_delay_table(inFile, dTable, delayData, iRow, &ierr);
      if (ierr != 0) {
         printf("ERROR:fits delay table read error %i row %li\n", ierr, iRow);
         return 1;
      }
      delayTimes[iRow-1]  = delayData->time;
      groupDelays[iRow-1] = clight * delayData->delay; /* meters */
      delayAmp[iRow-1]    = delayData->amplitude; 
   }
/*
      kill delay tables and close file;
*/
   kill_delay_table_def(dTable);
   kill_delay_table_row(delayData);
   fits_close_file(inFile, &ierr);

/*
 *       median smooth to remove jumps before interpolating
 *       */
   minrtsMedianSmooth(groupDelays, nDelay, 5, medianGroupDelay);
/*
        calculate jumps in groupOpd (from 2nd difference )and set flags if too large
*/
   for (iRow = 1; iRow <nDelay-1; iRow++){
      jOpd[iRow] = groupDelays[iRow] - 0.5 *(groupDelays[iRow-1] + groupDelays[iRow+1]);
      if (fabs(jOpd[iRow] > jumpOpd))  jFlags[iRow] = 1;
   }

/*
 *       now use 2nd differences to calculate jitterOpd
 *       first median smooth by 5 to reduce influence of single jumps
 *       then take absolute value and lastly boxsmooth down by a few times
 *       gsmooth
*/
;  minrtsMedianSmooth(jOpd, nDelay, 5, medianJitter);
;  for (iRow =0;iRow<nDelay;iRow++) medianJitter[iRow]=fabs(medianJitter[iRow]);
   nRflag   = 0;
   rmsRflag = 0.;

   for (iRow =0;iRow<nDelay;iRow++) jOpd[iRow] = 0.;
/*                 form 2nd difference of group delays 
 *                 then take absolute value
 *                 then boxcar smooth and compared to specified max
 *                 meanwhile accumulate rms jitter of non-flagged delaytimes
 *                 */
   for (iRow = jitterDd; iRow <nDelay-jitterDd; iRow++) /* don't include end points in estimates */
      jOpd[iRow] = groupDelays[iRow] - 0.5 *(groupDelays[iRow-jitterDd] + groupDelays[iRow+jitterDd]);
   for (iRow =0;iRow<nDelay;iRow++) medianJitter[iRow]=fabs(jOpd[iRow]);
   minrtsBoxSmooth(nDelay, medianJitter, jOpd, JITTERBOX);
   for (iRow = jitterDd; iRow <nDelay-jitterDd; iRow++)
      if (fabs(jOpd[iRow]) > jitterOpd)  rFlags[iRow] = 1; 
      else {
         nRflag++;
         rmsRflag = rmsRflag +jOpd[iRow]*jOpd[iRow];
      }
   if (nRflag > 0) rmsRflag = sqrt(rmsRflag/nRflag);
/*
         amplitude clipping
*/
   nAmp = 0;
   if ((dNoise > 0.) && (snLevel > 0)) {
      snLevel = snLevel * dNoise;
      for (iRow=0;iRow<nDelay;iRow++) if (delayAmp[iRow] < snLevel) 
         aFlags[iRow] = 1;
   }
/*
      open input file and get detector info
*/
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits openfile of %s gave error %i\n",argv[1],
         ierr);
      printf("%s .\n",strerror(errno));
      return 1;
   }
/*
      create output flag table and row and fill in constant values
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[3]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      printf ("ERROR: opening fits flag output %s %i\n",argv[3], ierr);
      return 1;
      }
/*
 *    copy primary header from input to output
*/
   i = fits_copy_header(inFile, outFile, &ierr);
/*    
      get grism setting and use to set opd diff limits
*/
   if (maxOpd == 0.) {
      i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", gris,
         0, &ierr);
      if (gris[0] == 'G') maxOpd = DELTAOPDGRISM;
         else maxOpd = DELTAOPDPRISM;
   }
   printf("   maxOpd = %5.1f mu, minOpd = %5.1f mu, jumpOpd = %5.1f mu, jitterOpd = %5.1f mu\n", 
      1.e6*maxOpd, 1.e6*minOpd, 1.e6*jumpOpd, 1.e6*jitterOpd );
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata       = open_imaging_fdata_table(inFile, &ierr);

/*     
   get row size to be used as "NO_CHAN" in flag table
*/
   nx        = DetectorDef->naxis[0][0];
   inData    = make_imaging_fdata_row(idata);
   i         = fits_get_num_rows (inFile,  &dataRows, &ierr);
   dFlags    = (int *)malloc(dataRows * sizeof(int));
   nanFlags  = (int *)malloc(dataRows * sizeof(int));
   dataTimes = (double *)malloc(dataRows * sizeof(double));
   ImOK      = (int *)malloc(dataRows * sizeof(int));


/* 
      now loop over input data, calculate OPD and compare to group delay
*/
   xLast = 0;
   for (iRow=1; iRow<=dataRows; iRow++) {
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
      if (!finite(inData->time)) {nanFlags[iRow-1] = 1; ImOK[iRow-1] = 0;}
      else {
      if (iRow ==1) {
         startDataTime = inData->time;
         for (i=0; i<nDelay;i++) delayTimes[i] = delayTimes[i] - startDataTime;
      }
      opd = inData->opd[1]+inData->local_opd[1] - inData->opd[0] - inData->local_opd[0];
      inTime = inData->time - startDataTime;
      dataTimes[iRow-1] = inTime;

      intOpd = quadInterpol(nDelay, delayTimes, medianGroupDelay, inTime, &xLast);

      dOpd = opd - intOpd;
      if (dOpd < 0.) dOpd = -dOpd;
      if (dOpd > maxOpd) {
         dFlags[iRow-1] = 1;
         ImOK[iRow-1]   = 0;
      } else if (dOpd < minOpd) {
         dFlags[iRow-1] = 1;
         ImOK[iRow-1]   = 0;
      } else {
         ImOK[iRow-1]   = 1; 
         dFlags[iRow-1] = 0;
      }
   }  /* checked whether time is finite */
   }  /* end of search for delta opd flagging */
   for (iRow=1;iRow<dataRows;iRow++) if(dFlags[iRow]) ++nMax;
/*
      shut down input data file
*/
   i = close_imaging_fdata_table(inFile, idata, &ierr);
   kill_imaging_fdata_row(inData);
/*
 *     it is double work, but let's mark flagged data in ImOK 
 *     now so we can calculate # of good points and 
 *     write it into header of output data
*/
   dRow = 0; /* counter for delay rows */
   sRow = 0; /* counter for data rows */
   while (dRow < nDelay) { /* loop over delay table Rows */
      if ((jFlags[dRow] == 0) && (rFlags[dRow]==0)&&(aFlags[dRow]==0))
         ++dRow;  /* there were not delay-related flags */
      else { /* now find the data rows between the flagged delay rows */
         time0 = delayTimes[dRow]; /* time of 1st flagged delay row */
/*                  loop over earlier data row */
         while(sRow < dataRows && (dataTimes[sRow] < time0)) sRow++;
/*                  now increment delay rows until you find a good one */
         while((jFlags[dRow] != 0)||(rFlags[dRow] != 0)||(aFlags[dRow] != 0) 
           && (dRow < nDelay)) ++dRow;
         time1 = delayTimes[dRow-1]; /* time at end of flagged delay info */
/*                  now loop over data rows between 1st and last flagged delay rows 
 *                  setting the ImOK counter as bad as we go along */
         while (sRow < dataRows && (dataTimes[sRow] <= time1)) {
            ImOK[sRow] = 0;
            sRow++;
         } 
      } 
   }
/* last 100 data rows are flagged */
   for (iRow=(dataRows-100);i<dataRows;i++) ImOK[iRow] = 0;
   nGood  = 0;
   for (iRow=0;iRow<dataRows;iRow++) nGood = nGood + ImOK[iRow];

   i = fits_write_key(outFile, TINT, "NGOOD", (void*)(&nGood), 
      "number of valid output points", &ierr);
   i = fits_write_key(outFile, TFLOAT, "RMSOPD", (void*)(&rmsRflag), 
      "rms OPD jitter(meters)", &ierr);
   snprintf(history, 80, "oirAutoFlag jumpOpd=%f", jumpOpd);
   i = fits_write_history(outFile, history, &ierr);
   snprintf(history, 80, "oirAutoFlag jitterOpd=%f", jitterOpd);
   i = fits_write_history(outFile, history, &ierr);
   snprintf(history, 80, "oirAutoFlag minOpd=%f", minOpd);
   i = fits_write_history(outFile, history, &ierr);
   snprintf(history, 80, "oirAutoFlag maxOpd=%f", maxOpd);
   i = fits_write_history(outFile, history, &ierr);
/*
 *       create table section of output
*/

   flagTable = make_flag_table_def (outFile, "1776-07-04", nx, &ierr); 
   if (ierr !=0) {
      printf ("ERROR: %i creating flag data definition\n", ierr);
      return 1;
      }
   flagData = make_flag_table_row(flagTable);
   flagData->sourceId     =  1;
   flagData->telescope[0] =  1;
   flagData->telescope[1] = -1;
   flagData->chans[0]     =  1;
   flagData->chans[1]     = -1;

/*
         now compress maxOpd data and output
*/ 
   oRow = 1;
   iRow = 0;
   strcpy(flagData->reason, "GROUP - TRACKING DELAY");
   while (iRow < dataRows) {
      if (dFlags[iRow] == 0) ++iRow; else {
         time0 = startDataTime + dataTimes[iRow];
         while(dFlags[iRow] != 0 && (iRow < dataRows)) ++iRow;
         time1 = startDataTime + dataTimes[iRow-1];
         flagData->timeRange[0] = time0;
         flagData->timeRange[1] = time1;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }
/*
 *           let's flag nan data
*/
   iRow = 1;
   strcpy(flagData->reason, "LCU TIMING ERROR");
   while (iRow < (dataRows-1)) {
      if (nanFlags[iRow] == 0) ++iRow; else {
         time0 = startDataTime + 0.5*(dataTimes[iRow-1]+dataTimes[iRow+1]);
         while(nanFlags[iRow] != 0 && (iRow < dataRows)) ++iRow;
         time1 = startDataTime + 0.5*(dataTimes[iRow-2]+dataTimes[iRow]);
         flagData->timeRange[0] = time0;
         flagData->timeRange[1] = time1;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }

   for (iRow=1;iRow<dataRows;iRow++) if(nanFlags[iRow]) ++nNan;
/*
         now compress jumpOpd data (stored in jFlags) and output
         note there is nothing against flagging data twice
         Use jFlags and iRow to figure out the times (stored
         in delayTimes) to be written in the flag record.  Then use
         sRow, which refers to the data times and not the delay
         times, to count how many frames in the time interval
         are flagged.  oRow counts the output record position.
*/ 
   iRow = 0;
   sRow = 0;
   strcpy(flagData->reason, "JUMP IN GROUP DELAY");
   while (iRow < nDelay) {
      if (jFlags[iRow] == 0) ++iRow; else {
         time0 = delayTimes[iRow];
         while(sRow < dataRows && (dataTimes[sRow] < time0)) sRow++;
         while(jFlags[iRow] != 0 && (iRow < nDelay)) ++iRow;
         time1 = delayTimes[iRow-1];
         while (sRow < dataRows && (dataTimes[sRow++] <= time1)) {
/*            ImOK[sRow++] = 0; */
            ++nJump;
         }
         flagData->timeRange[0] = time0 + startDataTime;
         flagData->timeRange[1] = time1 + startDataTime;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }
/*
         now compress jitterOpd data (stored in rFlags) and output
         note there is nothing against flagging data twice
*/ 
   iRow = 0;
   sRow = 0;
   strcpy(flagData->reason, "Excessive OPD jitter");
   while (iRow < nDelay) {
      if (rFlags[iRow] == 0) ++iRow; else {
         time0 = delayTimes[iRow];
         while(sRow < dataRows && (dataTimes[sRow] < time0)) sRow++;
         while(rFlags[iRow] != 0 && (iRow < nDelay)) ++iRow;
         time1 = delayTimes[iRow-1];
         while (sRow < dataRows && (dataTimes[sRow++] < time1)) {
/*            ImOK[sRow++] = 0; */
            ++nJit;
         }
         flagData->timeRange[0] = time0 + startDataTime;
         flagData->timeRange[1] = time1 + startDataTime;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }
/*
         now compress amplitude data (stored in aFlags) and output
*/ 
   iRow = 0;
   sRow = 0;
   strcpy(flagData->reason, "Low S/N");
   while (iRow < nDelay) {
      if (aFlags[iRow] == 0) ++iRow; else {
         time0 = delayTimes[iRow];
         while(sRow < dataRows && (dataTimes[sRow] < time0)) sRow++;
         while(aFlags[iRow] != 0 && (iRow < dataRows)) ++iRow;
         time1 = delayTimes[iRow-1];
         while (sRow < dataRows && (dataTimes[sRow] < time1)) {
 /*           ImOK[sRow++] = 0; */
            ++nAmp;
         }
         flagData->timeRange[0] = time0 + startDataTime;
         flagData->timeRange[1] = time1 + startDataTime;
         i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
         if (ierr != 0) {
            printf("ERROR:output error at line %li %i\n", oRow-1, ierr);
            return 1;
         }
      }
   }
/*
      cludge; flag last 100 rows
*/
  flagData->timeRange[0] = dataTimes[dataRows-100]+startDataTime;
  flagData->timeRange[1] = dataTimes[dataRows-1]+startDataTime;
  strcpy(flagData->reason, "END of OBS");
  i = write_flag_table(outFile, flagTable, flagData, oRow++, &ierr);
/*
      shut down and shut up
*/
   kill_flag_table_def(flagTable);
   kill_flag_table_row(flagData);
   fits_close_file(outFile, &ierr);
   free(delayTimes);
   free(dataTimes);
   free(groupDelays);
   free(medianGroupDelay);
   free(medianJitter);
   free(delayAmp);
   free(dFlags);
   free(nanFlags);
   free(jFlags);
   free(rFlags);
   free(aFlags);
   free(jOpd);
   free(ImOK);
   printf("   Out of %li input points flagged\n", dataRows);
   printf("      - %6li OPD jumps,\n", nJump);
   printf("      - %6li LCU timing errors,\n", nNan);
   printf("      - %6li noisy delay points,\n", nJit);
   printf("      - %6li points for excessive/insufficient OPD difference,\n", nMax);
   printf("      - %6li points for insufficient S/N.\n", nAmp);
   printf("   There remain %li good points.\n", nGood);
   printf("   RMS OPD jitter of good points is %5.2f microns.\n",rmsRflag*1.e6);
   printf("   Note that some of the above flagged points may have been counted twice,\n");
   printf("   e.g. an OPD jump may occur at an excessive OPD point.\n");
   return 0;
}
